#include <stdio.h>
#include <stdlib.h>
#include <SDL/SDL.h>
#include <math.h>
#include "types.h"
#include "polygon.h"

/*  advance the index by one vertex forward through the vertex list,
    wrapping at the end of the list */
#define INDEX_FORWARD(Index)    Index = (Index + 1) % VertexList->Length;

/*  advance the index by one vertex backward through the vertex list,
    wrapping at the start of the list */
#define INDEX_BACKWARD(Index)    Index = (Index - 1 + VertexList->Length) % VertexList->Length;

#define INDEX_MOVE(Index,Direction)                       \
            if(Direction > 0)                             \
            {                                             \
                Index = (Index + 1) % VertexList->Length; \
            }                                             \
            else                                          \
            {                                             \
                Index = (Index - 1 + VertexList->Length) % VertexList->Length; \
            };

extern SDL_Surface* scr;

static S32 ClipMinX = 0;
static S32 ClipMinY = 0;
static S32 ClipMaxX = SCREEN_WIDTH;
static S32 ClipMaxY = SCREEN_HEIGHT;

static void DrawHorizontalLineList(struct HLineList *HLineListPtr, S32 Colour);
static void ScanEdge(S32 X1, S32 Y1, S32 X2, S32 Y2, S32 SetXStart,
        S32 SkipFirst, struct HLine **EdgePointPtr);

S32 FillConvexPolygon(struct PointListHeader *VertexList ,S32 Colour,
                                 S32 XOffset, S32 YOffset)
{
    struct Point *VertexPtr;
    S32 MinIndexL;
    S32 MinIndexR;
    S32 MaxIndex;
    S32 MaxPoint_Y;
    S32 MinPoint_Y;
    S32 i, SkipFirst;
    S32 Temp;
    S32 NextIndex, PreviousIndex, CurrentIndex;
    S32 LeftEdgeDir;
    S32 TopIsFlat;
    S32 DeltaXN, DeltaYN, DeltaXP, DeltaYP;
    struct HLineList WorkingHlineList;
    struct HLine  *EdgePointPtr;

    /* point to the vertex list */
    VertexPtr = VertexList->PointPtr;

    /* scan the list to find top and bottom of the polygon */
    if(VertexList->Length == 0)
    {
        /* reject null polygons  */
        return(1);
    }

    MinIndexL = 0;
    MaxIndex = 0;
    MaxPoint_Y = MinPoint_Y = VertexPtr[0].Y;

    for(i = 1; i < VertexList->Length; i++)
    {
        if(VertexPtr[i].Y < MinPoint_Y)
        {
            MinIndexL = i;               /* new top */
            MinPoint_Y = VertexPtr[i].Y;
        }
        else if(VertexPtr[i].Y > MaxPoint_Y)
        {
            MaxIndex = i;                /* new bottom */
            MaxPoint_Y = VertexPtr[i].Y;
        }
    }

    /* check if polygon is 0 height: avoid infinite loop below */
    if(MinPoint_Y == MaxPoint_Y)
    {
        return(1);
    }

    /* scan in ascending order to find the last top-edge point */
    MinIndexR = MinIndexL;

    while(VertexPtr[MinIndexR].Y == MinPoint_Y)
    {
        INDEX_FORWARD(MinIndexR);
    }

    INDEX_BACKWARD(MinIndexR) /* back up to last top-edge point */

    /* Now scan in decending order to find the first top-edge point */
    while(VertexPtr[MinIndexL].Y == MinPoint_Y)
    {
        INDEX_BACKWARD(MinIndexL)
    }

    INDEX_FORWARD(MinIndexL) /* back up to first top-edge point */

    /* figure out which direction through the vertex list from the top */
    /* vertex is the left egde and which is the right                  */

    LeftEdgeDir = -1; /* assume left edge runs down thru vertex list */

    TopIsFlat = (VertexPtr[MinIndexL].X != VertexPtr[MinIndexR].X) ? 1:0;

    if(TopIsFlat == 1)
    {
        /* if the top is flat, just see which of the ends is leftmost */
        if(VertexPtr[MinIndexL].X > VertexPtr[MinIndexR].X)
        {
            LeftEdgeDir = 1;       /* left edge runs up through vertex list */
            Temp = MinIndexL;      /* swap the indices so MinIndexL         */
            MinIndexL = MinIndexR;  /* points to the start of the left       */
            MinIndexR = Temp;      /* edge, similarly for MinIndexR         */
        }
    }
    else
    {
        /* point to the downward end of the first line of each of the */
        /* two edges down from the top */
        NextIndex = MinIndexR;
        INDEX_FORWARD(NextIndex);
        PreviousIndex = MinIndexL;
        INDEX_BACKWARD(PreviousIndex);

        /* calculate X and Y lengths from the top vertex to the end of */
        /* the first line down each edge, use those to compare slopes  */
        /* and see which line is leftmost                              */
        DeltaXN = VertexPtr[NextIndex].X - VertexPtr[MinIndexL].X;
        DeltaYN = VertexPtr[NextIndex].Y - VertexPtr[MinIndexL].Y;
        DeltaXP = VertexPtr[PreviousIndex].X - VertexPtr[MinIndexL].X;
        DeltaYP = VertexPtr[PreviousIndex].Y - VertexPtr[MinIndexL].Y;
        if(((long)DeltaXN * DeltaYP - (long)DeltaYN * DeltaXP) < 0L)
        {
            LeftEdgeDir = 1;       /* left edge runs up through vertex list */
            Temp = MinIndexL;      /* swap the indices so MinIndexL         */
            MinIndexL = MinIndexR; /* points to the start of the left       */
            MinIndexR = Temp;      /* edge, similarly for MinIndexR         */
        }
    }

    /* Set the # of scan lines in the polygon, skipping the bottom edge */
    /* and also skipping the top vertex if the top isn't flat because   */
    /* in that case the top vertex has a right edge component, and set  */
    /* the top scan line to draw, which is likewise the second line of  */
    /* the polygon unless the top is flat */
    WorkingHlineList.Length = MaxPoint_Y - MinPoint_Y - 1 + TopIsFlat;
    if(WorkingHlineList.Length <= 0)
    {
        return(1); /* theres nothing to draw, so we're done */
    }

    WorkingHlineList.YStart = YOffset + MinPoint_Y + 1 - TopIsFlat;

    /* Get memory in which to store the line list we generate */
    WorkingHlineList.HLinePtr = (struct HLine *) malloc(sizeof(struct HLine) * WorkingHlineList.Length);
    if(WorkingHlineList.HLinePtr == NULL)
    {
        return(0);    /* couldnt get memory for the line list */
    }

    /* scan the left edge and store the boundary points in the list */
    /* initial pointer for storing scan converted left edge coords  */
    EdgePointPtr = WorkingHlineList.HLinePtr;
    /* start from the top of the left edge */
    PreviousIndex = CurrentIndex = MinIndexL;

    /* Skip the first point of the first line unless the top is flat    */
    /* if the top isn't flat, the top vertex is exactly on a right edge */
    /* and isn't drawn */
    SkipFirst = TopIsFlat ? 0 : 1;

    /* scan convert each line in the left edge from top to bottom */
    do
    {
        INDEX_MOVE(CurrentIndex,LeftEdgeDir);
        ScanEdge(VertexPtr[PreviousIndex].X + XOffset,
                VertexPtr[PreviousIndex].Y,
                VertexPtr[CurrentIndex].X + XOffset,
                VertexPtr[CurrentIndex].Y, 1, SkipFirst, &EdgePointPtr);
        PreviousIndex = CurrentIndex;
        SkipFirst = 0; /* scan convert the first point from now on */
    }while(CurrentIndex != MaxIndex);

    /* Scan the right edge and store the boundary points in the list */
    EdgePointPtr =  WorkingHlineList.HLinePtr;
    PreviousIndex = CurrentIndex = MinIndexR;
    SkipFirst = TopIsFlat ? 0 : 1;
    /* Scan convert the right edge, top to bottom, X coordinates are  */
    /* adjusted 1 to the left, effectively causing scan conversion of */
    /* the nearest points to the left of but not exactly on the edge  */
    do
    {
        INDEX_MOVE(CurrentIndex,-LeftEdgeDir);
        ScanEdge(VertexPtr[PreviousIndex].X + XOffset - 1,
                VertexPtr[PreviousIndex].Y,
                VertexPtr[CurrentIndex].X + XOffset - 1,
                VertexPtr[CurrentIndex].Y, 0, SkipFirst, &EdgePointPtr);
        PreviousIndex = CurrentIndex;
        SkipFirst = 0; /* scan convert the first point from now on */
    }while(CurrentIndex != MaxIndex);

    /* draw the line list representing the scan converted polygon */
    DrawHorizontalLineList(& WorkingHlineList, Colour);

    /* Release the line list's memory and we're successfully done */
    free( WorkingHlineList.HLinePtr);
    return(1);
}

/* Scan convert an edge from (X1,Y1) to (X2,Y2), not including the point  */
/* at (X2,Y2). This avoids overlapping the end of one line with the start */
/* of the next, and causes the bottom scan line of the polygon not to be  */
/* drawn. If SkipFirst != 0, the point at (X1,Y1) isnt drawn. For each    */
/* scan line, the pixel closest to the scanned line without being to the  */
/* left of the scanned line is choosen                                    */
static void ScanEdge(S32 X1, S32 Y1, S32 X2, S32 Y2, S32 SetXStart,
        S32 SkipFirst, struct HLine **EdgePointPtr)
{
    struct HLine *WorkingEdgePointPtr;
    S32 AdvanceAmt;
    S32 DeltaX;
    S32 Height;
    S32 Width;
    S32 i;
    S32 ErrorTerm;
    S32 XMajorAdvanceAmt;
    S32 ErrorTermAdvance;

    WorkingEdgePointPtr = *EdgePointPtr; /* avoid double dereference */
    
    DeltaX = X2 - X1; 
    /* DIRECTION IN WHICH x MOVES (y2 IS ALWAYS > y1, SO y ALWAYS COUNTS UP) */
    AdvanceAmt = (DeltaX > 0) ? 1 : -1;
    
    Height = Y2 - Y1;
    if(Height <= 0)
    {
        return;    /* guard against 0 length and horizontal edges */          
    }
    
    /* Figure out whether the edge is vertical, diagonal, X-major  */
    /* (mostly horizontal) or Y-major (mostly vertical) and handle */
    /* appropriately                                               */
    Width = abs(DeltaX);
    
    if(Width == 0)
    {
        /* The edge is vertical: special-case by just storing the same */
        /* X coordinate for every scan line                            */
        
        /* Scan the edge for each scan line in turn */
        for(i = Height - SkipFirst; i-- > 0; WorkingEdgePointPtr++)
        {
              if(SetXStart == 1)
              {
                  WorkingEdgePointPtr->Xstart = X1;             
              }
              else
              {
                  WorkingEdgePointPtr->Xend = X1; 
              }                 
        }
    }
    else if(Width == Height)
    {
        /* The edge is diagonal: special-case by advancing the X */
        /* coordinates 1 pixel for each scan line */
        if(SkipFirst == 1) /* skip the fitst point if so indicated */
        {
            X1 += AdvanceAmt;    /* move 1 pixel to the left or right */              
        }
        
        /* scan the edge for each scan line in turn */
        for(i = Height - SkipFirst; i-- > 0; WorkingEdgePointPtr++)
        {
            /* store the X coordinates in the appropiate edge list */
            if(SetXStart == 1)
            {
                 WorkingEdgePointPtr->Xstart = X1;
            }
            else
            {
                 WorkingEdgePointPtr->Xend = X1;   
            }
            
             X1 += AdvanceAmt;    /* move 1 pixel to the left or right */            
        }   
    }
    else if(Height > Width)
    {
        /* edge is closer to vertical than horizontal (Y-major) */
        if(DeltaX >= 0)
        {
            ErrorTerm = 0; /* inital error term going left > right */          
        }
        else
        {
            ErrorTerm = -Height + 1; /* going right > left */
        }
        
        /* skip the first point if so indicated */
        if(SkipFirst == 1)
        {
            /* determine whether its time for the X coord to advance */
            ErrorTerm += Width;
            if(ErrorTerm > 0)
            {
                X1 += AdvanceAmt;    /* move 1 pixel to the left or right */ 
                ErrorTerm -= Height; /* *advance ErrorTerm to next point */               
            }                              
        }
                
        /* Scan the edge for each scan line in turn */
        for(i = Height - SkipFirst; i-- > 0 ; WorkingEdgePointPtr++)
        {
            /* store the X coordinates in the appropriate edge list */
            if(SetXStart == 1)
            {
                WorkingEdgePointPtr->Xstart = X1;
            }
            else
            {
                WorkingEdgePointPtr->Xend = X1;
            }
            
            /* Determine whether its time for the X coord to advance */
            ErrorTerm += Width;
            if(ErrorTerm > 0)
            {
                X1 += AdvanceAmt;    /* move 1 pixel to the left or right */ 
                ErrorTerm -= Height; /* *advance ErrorTerm to next point */               
            }                             
        }
    }
    else
    {
        /* Egde is closer to horizontal than vertical (X-major) */
            
        /* minimum distance to advance X each time */
        XMajorAdvanceAmt = (Width / Height) * AdvanceAmt;
        /* Error term advance for deciding when to advance X 1 extra */
        ErrorTermAdvance = Width % Height;
        if(DeltaX >= 0)
        {
            ErrorTerm = 0; /* initial error term going left > right */
        }
        else
        {
            ErrorTerm = -Height + 1; /* going right > left */    
        }
        
         /* skip the first point if so indicated */
        if(SkipFirst == 1)
        {
            X1 += XMajorAdvanceAmt; /* move X minimum distance */
                     
            /* determine whether its time for the X coord to advance */
            ErrorTerm +=ErrorTermAdvance;
            if(ErrorTerm > 0)
            {
                X1 += AdvanceAmt;    /* move 1 pixel to the left or right */ 
                ErrorTerm -= Height; /* *advance ErrorTerm to next point */               
            }                              
        }
        
        /* Scan the edge for each scan line in turn */
        for(i = Height - SkipFirst; i-- > 0 ; WorkingEdgePointPtr++)
        {
            /* store the X coordinates in the appropriate edge list */
            if(SetXStart == 1)
            {
                WorkingEdgePointPtr->Xstart = X1;
            }
            else
            {
                WorkingEdgePointPtr->Xend = X1;
            }
            
            X1 += XMajorAdvanceAmt; /* move X minimum distance */
            
            /* determine whether its time for the X coord to advance */
            ErrorTerm +=ErrorTermAdvance;
            if(ErrorTerm > 0)
            {
                X1 += AdvanceAmt;    /* move 1 pixel to the left or right */ 
                ErrorTerm -= Height; /* *advance ErrorTerm to next point */               
            }  
        }
    }
    
    *EdgePointPtr = WorkingEdgePointPtr; /* advance caller's ptr */
}

/* Draws all pixels in the list of horizontal lines passed in */
static void DrawHorizontalLineList(struct HLineList *HLineListPtr, S32 Colour)
{
    struct HLine *HLinePtr;
    S32 Y;
    S32 X;
    U32* bufp;
    U32 ColorSDL;
    color_type ColourUsed;
    
    /* point to the xstart / xend descriptor for the first(top) horz line */
    HLinePtr = HLineListPtr->HLinePtr;
    
    /* do we have any lines to draw ? */   
    if(HLineListPtr->Length <= 0)
    {
        return;    /* no then return */                        
    }
    
    /* do we need to clip the top ? */
    if(HLineListPtr->YStart < ClipMinY)
    {
        HLineListPtr->Length -= ClipMinY - HLineListPtr->YStart;
        /* do we have any lines to draw ? */   
        if(HLineListPtr->Length <= 0)
        {
            return;    /* no then return */                        
        }
        
        HLinePtr += ClipMinY - HLineListPtr->YStart;
        HLineListPtr->YStart += ClipMinY - HLineListPtr->YStart;                                
    }
    
    /* do we need to clip the bottom */
    if((HLineListPtr->YStart + HLineListPtr->Length) >= ClipMaxY)
    {
        HLineListPtr->Length -= (HLineListPtr->YStart + HLineListPtr->Length) - ClipMaxY;
         /* do we have any lines to draw ? */   
        if(HLineListPtr->Length <= 0)
        {
            return;    /* no then return */                        
        }                     
    }
    
    switch(Colour)
    {
      /* black */          
      case 0:
          ColourUsed.r = 0;
          ColourUsed.g = 0;
          ColourUsed.b = 0;
      break;          
                
      /* Blue */
      case 1:
           ColourUsed.r = 0;
           ColourUsed.g = 0;
           ColourUsed.b = 170;
      break;
                
      /* Green  */          
      case 2:
           ColourUsed.r = 0;
           ColourUsed.g = 170;
           ColourUsed.b = 0;
      break;
      
      /* cyan */          
      case 3:
          ColourUsed.r = 0;
          ColourUsed.g = 170;
          ColourUsed.b = 170;
      break;
      
      /* Red  */     
      case 4: 
          ColourUsed.r = 170;
          ColourUsed.g = 0;
          ColourUsed.b = 0;        
      break;
      
       /* MAGENTA  */     
      case 5: 
          ColourUsed.r = 170;
          ColourUsed.g = 0;
          ColourUsed.b = 170;        
      break;     
                  
      /* brown */
      case 6:
          ColourUsed.r = 170;
          ColourUsed.g = 85;
          ColourUsed.b = 0;
      break;
      
      /* LIGHT GRAY */
      case 7:
          ColourUsed.r = 170;
          ColourUsed.g = 170;
          ColourUsed.b = 170;
      break;
      
      /* black */
      case 40:
          ColourUsed.r = 0;
          ColourUsed.g = 0;
          ColourUsed.b = 0;
      break;
      
      default:
          /* do nothing */
      break;                             
    } 
    
    /* draw each horz line in turn, starting with the top one and advancing */
    /* one line each time */  
       
    ColorSDL = SDL_MapRGB(scr->format, ColourUsed.r, ColourUsed.g,ColourUsed.b);
       
    for(Y = HLineListPtr->YStart; Y < HLineListPtr->YStart + HLineListPtr->Length; Y++, HLinePtr++)
    {      
        /* do we need to clip the left edge */
        if(HLinePtr->Xstart < ClipMinX)
        {
            HLinePtr->Xstart = ClipMinX;
        }
        
        /* do we need to clip the right edge */
        if(HLinePtr->Xend >= ClipMaxX)
        {
            HLinePtr->Xend = ClipMaxX;
        }
        
        bufp = (U32*)scr->pixels + Y * scr->pitch / 4 + HLinePtr->Xstart;   
          
        /* Draw each pixel in the current horz line in turn starting with the
           leftmost one  */
        for(X = HLinePtr->Xstart; X <= HLinePtr->Xend; X++)
        {
            *bufp = ColorSDL;
            bufp++;
        }
    }
}
