#include <stdio.h>
#include <stdlib.h>
#include <SDL/SDL.h>
#include <math.h>
#include "types.h"
#include "polygon.h"
#include "vl.h"

#define TEST_POLY_LENGHT (sizeof(TestPoly)/sizeof(struct Point3))

/* transformation from polygons object space to world space.          */
/* Initially set up to perform no rotation and to move the polygon    */
/* into world space -140 units away from the origin down the z axis.  */
/* given the viewing point, -140 down the z axis means 140 units away */
/* straight ahead in the direction of view. The program dynamically   */
/* changes the rotation and translation                               */
static double PolyWorldXform[4][4] =
{
    {1.0,0.0,0.0,0.0},
    {0.0,1.0,0.0,0.0},
    {0.0,0.0,1.0,-140.0},
    {0.0,0.0,0.0,1.0}       
};
/* Transformation from world space into view space. In this program,   */
/* the view point is fixed at the origin of the world space. looking   */
/* down the z-axis in the direction of increasingly negative Z, so     */
/* view space is identical to world space; this is the identity matrix */
static double WorldViewXform[4][4]=
{
    {1.0,0.0,0.0,0.0},
    {0.0,1.0,0.0,0.0},
    {0.0,0.0,1.0,0.0},
    {0.0,0.0,0.0,1.0}        
};

int main(int argc, char *argv[])
{
    double WorkingXform[4][4];
    static struct Point3 TestPoly[] = {{-30,-15,0,1},{0,15,0,1},{10,-5,0,1}};
    double Rotation = M_PI / 60.0;    /* initial rotation = 3 degrees */
    SDL_Event event = {0};
    input_type UserControls;
    memset(&UserControls,0,sizeof(UserControls));
    
    /* try and get a handle to the screen */
    if(InitSDLScreenBuf() == 1)
    {
        /* cant get handle to screen so exit */
        return(1);
    }
    
    /* keep rotating the polygon */
    while(UserControls.Exit == 0)
    {
         /* clear last user input */
         memset(&UserControls,0,sizeof(UserControls));
         /* then see if user was pressed a key */
         GetUserInput(&UserControls);
         
         /* left (-x) */ 
         if(UserControls.Left == 1)
         {
             PolyWorldXform[0][3] -= 3.0;                         
         }
         /* right (+x) */
         if(UserControls.Right == 1)
         {
             PolyWorldXform[0][3] += 3.0;                         
         }
         /* up (+y) */
         if(UserControls.Forward == 1)
         {
             PolyWorldXform[1][3] += 3.0;                         
         }
         /* down (-y) */
         if(UserControls.Back == 1)
         {
             PolyWorldXform[1][3] -= 3.0;                         
         }                       
                            
         /* Modify the object space to world space transformation matrix */
         /* for the current rotation around the Y axis                   */
         PolyWorldXform[0][0] = PolyWorldXform[2][2] = cos(Rotation);
         PolyWorldXform[2][0] = -(PolyWorldXform[0][2] = sin(Rotation));
         
         /* Concatenate the object-to-world and world-to-view            */
         /* transformations to make a transformation matrix that will    */
         /* convert vertices from object space to view space in a single */
         /* operation                                                    */
         ConcatXforms(WorldViewXform, PolyWorldXform, WorkingXform);
         
         /* Transform the polygon, project it on the screen, draw it     */
         XformAndProjectPoly(WorkingXform, TestPoly, TEST_POLY_LENGHT, 4);
         
         /* Rotate 6 degrees farther around the Y axis                   */
         Rotation += M_PI / 30.0;
         if(Rotation >= M_PI * 2)
         {
             Rotation -= M_PI * 2;             
         }
    
         /* update screen */
         Redraw();
            
         /* clear screen buffer for next write */
         Cls();

         SDL_Delay(50); /* so it consumes less processing power */                  
    };
}
