#ifndef POLY_H
#define POLY_H

#define MAX_POLY_LENGTH    4   /* four vertices is the max per poly */

#define SCREEN_WIDTH    320
#define SCREEN_HEIGHT   240

/* Ratio: distance from viewpoint to projection plane / width of   */
/* projection plane. Defines the width of the field of view. Lower */
/* absolute value = wider fields of view; higher values = narrower */
#define PROJECTION_RATIO -2.0 /* negative because visible Z coors are negative */

/* Draws the polygon described by the point list PointList in colour */
/* Colour with all vertices offset by (X,Y) */
#define DRAW_POLYGON(PointList, NumPoints, Colour, X, Y)       \
    Polygon.Length = NumPoints;                                \
    Polygon.PointPtr = PointList;                              \
    FillConvexPolygon(&Polygon, Colour, X, Y);

/* describes a single point (used for a single vertex) */
struct Point
{
    S32 X; /* X coordinate */
    S32 Y; /* Y coordinate */
};

/* Describes a single 3-d point in homogeneous coordinates */
struct Point3 
{
    double X;    /* X coordinate */
    double Y;    /* Y coordinate */
    double Z;    /* X coordinate */
    double W;       
};


/* describes a series of point (used to store a list of vertices that
   describe a polygon, each vertex is assumed to connect to the two
   adjacent vertices, and the last vertex is assumed to connect to the
   first) */
struct PointListHeader
{
    S32 Length;             /* # of points */
    struct Point *PointPtr; /* pointer to list of points */
};

/* describes the beginning and ending X coordinates of a single
   horizontal line */
struct HLine
{
    S32 Xstart;  /* X coord of the leftmost pixel in the line */
    S32 Xend;    /* X coord of the rightmost pixel in the line */
};

/* describes a length-long series of horizontal lines, all assumed to be on
   contiguous scan lines starting at Ystart and proceeding downward (used
   to describe a scan-converted polygon to the low-level hardware-dependent
   drawing code) */
struct HLineList
{
    S32 Length;    /* # of horizontal lines */
    S32 YStart;    /* Y coord of topmost line */
    struct HLine *HLinePtr; /* pointer to list of horz lines */
};

#endif
